/*
** nester - NES emulator
** Copyright (C) 2000  Darren Ranalli
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful, 
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
** Library General Public License for more details.  To obtain a 
** copy of the GNU Library General Public License, write to the Free 
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
*/

#ifndef NES_SETTINGS_H_
#define NES_SETTINGS_H_

#include "types.h"

#include "OSD_NES_graphics_settings.h"
#include "OSD_ButtonSettings.h"
#include "nnnextkeycfg.h"

class NES_preferences_settings
{
public:
  boolean run_in_background;
  boolean SkipSomeErrors;
  boolean speed_throttling;
  boolean auto_frameskip;
  int FastFPS;
  boolean ToggleFast;
  enum NES_PRIORITY { PRI_NORMAL=0, PRI_HIGH=1, PRI_REALTIME=2 };
  NES_PRIORITY priority;

  unsigned char UseRomDataBase;										//nnn
  unsigned char AutoRomCorrect;
  unsigned char AutoStateSL;
  unsigned char AutoStateSLNotload;
  unsigned char AllowMultiInstance;
  unsigned char ZipDLLUseFlag;

  void SetDefaults()
  {
    run_in_background = FALSE;
	SkipSomeErrors = BST_CHECKED;
    speed_throttling = TRUE;
    auto_frameskip = TRUE;
    priority = PRI_HIGH;
	FastFPS = 9;
	ToggleFast = TRUE;
	UseRomDataBase=0;										//nnn
	AutoRomCorrect=0;
	AutoStateSL=0;
	AutoStateSLNotload=0;
	AllowMultiInstance=0;
	ZipDLLUseFlag=0;
  }

  NES_preferences_settings()
  {
    SetDefaults();
  }
};

class NES_graphics_settings
{
public:
  boolean black_and_white;
  boolean show_more_than_8_sprites;
  boolean show_all_scanlines;
  boolean draw_overscan;
  boolean fullscreen_on_load;
//  boolean fullscreen_scaling;
  uint8 fullscreen_scaling;		//nnn change
  boolean calculate_palette;
  boolean UseStretchBlt;
  boolean EmulateTVScanline;
  char szPaletteFile[260];
  uint8 tint;
  uint8 hue;
  OSD_NES_graphics_settings osd;

  uint8 WindowModeUseDdraw;
  uint8 WindowModeUseDdrawOverlay;
  uint8 FullscreenModeUseNewMode;
  uint8 FullscreenModeUseDefRefreshRate;
  uint8 FullscreenModeBackBuffM;
  uint8 WindowModeBackBuffM;
  uint8 FullscreenModeNotUseFlip;

  uint8 WindowTopMost;
  uint8 ScreenShotUseBmp;

  void reset_palette()
  {
    tint = 0x86;
    hue  = 0x9d;
  }

  void SetDefaults()
  {
    black_and_white = FALSE;
    show_more_than_8_sprites = FALSE;
    show_all_scanlines = FALSE;
    draw_overscan = FALSE;
    fullscreen_on_load = FALSE;
    fullscreen_scaling = FALSE;
    calculate_palette = FALSE;
	UseStretchBlt = FALSE;
	EmulateTVScanline = FALSE;
	WindowModeUseDdraw=0;
	WindowModeUseDdrawOverlay=0;
	FullscreenModeUseNewMode=0;
	FullscreenModeUseDefRefreshRate=0;
	FullscreenModeBackBuffM=0;
	WindowModeBackBuffM=0;
	WindowTopMost=0;
	ScreenShotUseBmp=0;
	FullscreenModeNotUseFlip=0;
    *szPaletteFile = '\0';
    reset_palette();
    osd.Init();
  }

  NES_graphics_settings()
  {
    SetDefaults();
  }
};

class NES_sound_settings
{
public:
  boolean enabled;
  uint32 sample_bits;
  uint32 sample_rate;

  boolean rectangle1_enabled;
  boolean rectangle2_enabled;
  boolean triangle_enabled;
  boolean noise_enabled;
  boolean dpcm_enabled;
  boolean ext_enabled;

  boolean ideal_triangle_enabled;
  boolean smooth_envelope_enabled; // reserved
  boolean smooth_sweep_enabled; // reserved


  float rectangle1_volumed;
  float rectangle2_volumed;
  float triangle_volumed;
  float noise_volumed;
  float dpcm_volumed;
  float ext_volumed;

  enum { LENGTH_MIN = 1, LENGTH_MAX = 10 };
  uint32 buffer_len;

  enum filter_type_t { FILTER_NONE, FILTER_LOWPASS, FILTER_LOWPASS_WEIGHTED };
  filter_type_t filter_type;

  void SetDefaults()
  {
    enabled = TRUE;
    sample_bits = 8;
    sample_rate = 44100;
    buffer_len = 3;

    filter_type = FILTER_LOWPASS_WEIGHTED;

    rectangle1_enabled = TRUE;
    rectangle2_enabled = TRUE;
    triangle_enabled = TRUE;
    noise_enabled = TRUE;
    dpcm_enabled = TRUE;
	ext_enabled = TRUE;

    ideal_triangle_enabled = FALSE;
    smooth_envelope_enabled = FALSE;
    smooth_sweep_enabled = FALSE;

	rectangle1_volumed=0;
	rectangle2_volumed=0;
	triangle_volumed=0;
	noise_volumed=0;
	dpcm_volumed=0;
	ext_volumed=0;
  }

  NES_sound_settings()
  {
    SetDefaults();
  }
};

class NES_controller_input_settings
{
public:
  OSD_ButtonSettings btnUp;
	OSD_ButtonSettings btnDown;
	OSD_ButtonSettings btnLeft;
	OSD_ButtonSettings btnRight;
	OSD_ButtonSettings btnSelect;
	OSD_ButtonSettings btnStart;
	OSD_ButtonSettings btnB;
	OSD_ButtonSettings btnA;

	OSD_ButtonSettings btnTB;	//nnn Auto fire button
	OSD_ButtonSettings btnTA;
	int btnTBsec;
	int btnTAsec;

  // OS-specific
  void OSD_SetDefaults(int num); // 0 == first player

  void Clear()
  {
    btnUp.Clear();
    btnDown.Clear();
    btnLeft.Clear();
    btnRight.Clear();
    btnSelect.Clear();
    btnStart.Clear();
    btnB.Clear();
    btnA.Clear();
    btnTB.Clear();
    btnTA.Clear();
  }

  NES_controller_input_settings(int num)
  {
    Clear();
    OSD_SetDefaults(num);
  }
};

class NES_input_settings
{
public:
  NES_controller_input_settings player1;
  NES_controller_input_settings player2;
  NES_controller_input_settings player3;
  NES_controller_input_settings player4;

  OSD_ButtonSettings extkeycfg[EXTKEYALLNUM];		//nnn
  void OSD_EKey_SetDefaults();
  void OSD_EKey_SetDefaults2();
  void ExtBtnClear()
  {
	  for(int i=0; i<EXTKEYALLNUM; ++i){
		  extkeycfg[i].Clear();
	  }
  }
  unsigned char extkeycfgFastFPSToggle;

  void SetDefaults()
  {
    player1.Clear();
    player2.Clear();
    player3.Clear();
    player4.Clear();
    player1.OSD_SetDefaults(0);
    player2.OSD_SetDefaults(1);
    player3.OSD_SetDefaults(2);
    player4.OSD_SetDefaults(3);
	ExtBtnClear();
	OSD_EKey_SetDefaults();
	OSD_EKey_SetDefaults2();
  }

  NES_input_settings() : player1(0), player2(1), player3(2), player4(3)
  {
	ExtBtnClear();
	OSD_EKey_SetDefaults();
	OSD_EKey_SetDefaults2();
  }
};

class NES_settings
{
public:
  NES_preferences_settings  preferences;
  NES_graphics_settings     graphics;
  NES_sound_settings        sound;
  NES_input_settings        input;

  NES_settings() : preferences(), graphics(), sound(), input()
  {
  }
};

#endif
